@@ -1,6 +1,6 @@
 /*
   zip_filerange_crc.c -- compute CRC32 for a range of a file
-  Copyright (C) 2008-2016 Dieter Baron and Thomas Klausner
+  Copyright (C) 2008-2019 Dieter Baron and Thomas Klausner
 
   This file is part of libzip, a library to manipulate ZIP archives.
   The authors can be contacted at <libzip@nih.at>
@@ -32,14 +32,13 @@
 */
 
 
-#include <stdio.h>
-
 #include "zipint.h"
 
 
 int
 _zip_filerange_crc(zip_source_t *src, zip_uint64_t start, zip_uint64_t len, uLong *crcp, zip_error_t *error) {
-    Bytef buf[BUFSIZE];
+    DEFINE_BYTE_ARRAY(buf, BUFSIZE);
+
     zip_int64_t n;
 
     *crcp = crc32(0L, Z_NULL, 0);
@@ -54,14 +53,21 @@ _zip_filerange_crc(zip_source_t *src, zip_uint64_t start, zip_uint64_t len, uLon
 	return -1;
     }
 
+    if (!byte_array_init(buf, BUFSIZE)) {
+	zip_error_set(error, ZIP_ER_MEMORY, 0);
+	return -1;
+    }
+
     while (len > 0) {
 	n = (zip_int64_t)(len > BUFSIZE ? BUFSIZE : len);
 	if ((n = zip_source_read(src, buf, (zip_uint64_t)n)) < 0) {
 	    _zip_error_set_from_source(error, src);
+	    byte_array_fini(buf);
 	    return -1;
 	}
 	if (n == 0) {
 	    zip_error_set(error, ZIP_ER_EOF, 0);
+	    byte_array_fini(buf);
 	    return -1;
 	}
 
@@ -70,5 +76,7 @@ _zip_filerange_crc(zip_source_t *src, zip_uint64_t start, zip_uint64_t len, uLon
 	len -= (zip_uint64_t)n;
     }
 
+    byte_array_fini(buf);
+
     return 0;
 }
