@@ -8,6 +8,7 @@ use File::Copy;
 use File::Path qw(mkpath remove_tree);
 use Getopt::Long qw(:config posix_default bundling no_ignore_case);
 use IPC::Open3;
+#use IPC::Cmd qw(run);
 use Storable qw(dclone);
 use Symbol 'gensym';
 use UNIVERSAL;
@@ -78,11 +79,8 @@ use UNIVERSAL;
 #	mkdir MODE NAME
 #	    create directory NAME with permissions MODE.
 #
-#	pipefile FILE
-#	    pipe FILE to program's stdin.
-#
-#	pipein COMMAND ARGS ...
-#	    pipe output of running COMMAND to program's stdin.
+#   precheck COMMAND ARGS ...
+#		if COMMAND exits with non-zero status, skip test.
 #
 #	preload LIBRARY
 #	    pre-load LIBRARY before running program.
@@ -104,6 +102,12 @@ use UNIVERSAL;
 #       stderr-replace REGEX REPLACEMENT
 #           run regex replacement over expected and got stderr output.
 #
+#	stdin TEST
+#	    Provide TEXT to program's stdin.
+#
+#	stdin-file FILE
+#	    pipe FILE to program's stdin.
+#
 #	stdout TEXT
 #	    program is expected to print TEXT to stdout.  If multiple
 #	    stdout commands are used, the messages are expected in
@@ -157,14 +161,15 @@ sub new {
 		'file-del' => { type => 'string string' },
 		'file-new' => { type => 'string string' },
 		mkdir => { type => 'string string' },
-		pipefile => { type => 'string', once => 1 },
-		pipein => { type => 'string', once => 1 },
+		precheck => { type => 'string...' },
 		preload => { type => 'string', once => 1 },
 		program => { type => 'string', once => 1 },
 		'return' => { type => 'int', once => 1, required => 1 },
 		setenv => { type => 'string string' },
 		stderr => { type => 'string' },
 		'stderr-replace' => { type => 'string string' },
+		stdin => { type => 'string' },
+		'stdin-file' => { type => 'string', once => 1 },
 		stdout => { type => 'string' },
 		touch => { type => 'int string' },
 		ulimit => { type => 'char string' }
@@ -340,7 +345,11 @@ sub runtest_one {
                 $preload_env_var = 'DYLD_INSERT_LIBRARIES';
         }
 	if (defined($self->{test}->{'preload'})) {
-		$ENV{$preload_env_var} = cwd() . "/../.libs/$self->{test}->{'preload'}";
+		if (-f cwd() . "/../.libs/$self->{test}->{'preload'}") {
+			$ENV{$preload_env_var} = cwd() . "/../.libs/$self->{test}->{'preload'}";
+		} else {
+			$ENV{$preload_env_var} = cwd() . "/../lib$self->{test}->{'preload'}";
+		}
 	}
 
 	$self->run_program();
@@ -352,6 +361,10 @@ sub runtest_one {
 		delete ${ENV{$preload_env_var}};
 	}
 
+	if ($self->{test}->{stdin}) {
+		$self->{stdin} = [ @{$self->{test}->{stdin}} ];
+	}
+
 	if ($self->{test}->{stdout}) {
 		$self->{expected_stdout} = [ @{$self->{test}->{stdout}} ];
 	}
@@ -411,6 +424,7 @@ sub setup {
 	@ARGV = @argv;
 	my $ok = GetOptions(
 		'help|h' => \my $help,
+	        'bin-sub-directory=s' => \$self->{bin_sub_directory},
 		'keep-broken|k' => \$self->{keep_broken},
 		'no-cleanup' => \$self->{no_cleanup},
 		# 'run-gdb' => \$self->{run_gdb},
@@ -421,7 +435,7 @@ sub setup {
 	@ARGV = @save_argv;
 
 	if (!$ok || scalar(@argv) != 1 || $help) {
-		print STDERR "Usage: $0 [-hv] [--keep-broken] [--no-cleanup] [--setup-only] testcase\n";
+		print STDERR "Usage: $0 [-hv] [--bin-sub-directory DIR] [--keep-broken] [--no-cleanup] [--setup-only] testcase\n";
 		exit(1);
 	}
 
@@ -439,8 +453,10 @@ sub setup {
 	$self->die("error in test case definition") unless $self->parse_case($testcase_file);
 
 	$self->check_features_requirement() if ($self->{test}->{features});
+	$self->run_precheck() if ($self->{test}->{precheck});
 
-	$self->end_test('SKIP') if ($self->{test}->{preload} && $^O eq 'darwin');
+	$self->end_test('SKIP') if ($self->{test}->{preload} && ($^O eq 'darwin' || $^O eq 'MSWin32'));
+	$self->end_test('SKIP') if ($self->{test}->{stdin} || $self->{test}->{'stdin-file'} && $^O eq 'MSWin32');
 }
 
 
@@ -499,7 +515,8 @@ sub check_features_requirement() {
 sub comparator_zip {
 	my ($self, $got, $expected) = @_;
 
-	my @args = ($self->{zipcmp}, $self->{verbose} ? '-v' : '-q');
+	my $zipcmp = (-f $self->{zipcmp}) ? $self->{zipcmp} : $self->find_program('zipcmp');
+	my @args = ($zipcmp, $self->{verbose} ? '-v' : '-q');
 	push @args, $self->{zipcmp_flags} if ($self->{zipcmp_flags});
 	push @args, ($expected, $got);
 
@@ -589,7 +606,7 @@ sub list_files {
 
 	while (scalar(@dirs) > 0) {
 		my $dir = shift @dirs;
-		
+
 		opendir($ls, $dir);
 		unless ($ls) {
 			# TODO: handle error
@@ -599,7 +616,7 @@ sub list_files {
 			if ($dir eq '.') {
 				$file = $entry;
 			}
-			
+
 			if (-f $file) {
 				push @files, "$file";
 			}
@@ -730,7 +747,7 @@ sub find_file() {
 
 	for my $dir (('', "$self->{srcdir}/")) {
 		my $f = "$dir$fname";
-		$f = "../$f" if ($self->{in_sandbox} && $dir !~ m,^/,);
+		$f = "../$f" if ($self->{in_sandbox} && $dir !~ m,^(\w:)?/,);
 
 		return $f if (-f $f);
 	}
@@ -772,7 +789,7 @@ sub get_variable {
 		}
 		close ($fh);
 	}
-	if (!defined($self->{$name} || $self->{$name} eq '')) {
+	if (!defined($self->{$name}) || $self->{$name} eq '') {
 		$self->die("cannot get variable $name");
 	}
 }
@@ -890,7 +907,7 @@ sub parse_case() {
 	my %test = ();
 
 	while (my $line = <TST>) {
-		chomp $line;
+		$line =~ s/(\n|\r)//g;
 
 		next if ($line =~ m/^\#/);
 
@@ -939,8 +956,8 @@ sub parse_case() {
 		}
 	}
 
-	if ($test{pipefile} && $test{pipein}) {
-		$self->warn_file("both pipefile and pipein set, choose one");
+	if ($test{'stdin-file'} && $test{stdin}) {
+		$self->warn_file("both stdin-file and stdin provided, choose one");
 		$ok = 0;
 	}
 
@@ -1081,9 +1098,9 @@ sub run_hook {
 
 	return $ok;
 }
+
+
 sub args_decode {
-
-
 	my ($str, $srcdir) = @_;
 
 	if ($str =~ m/\\/) {
@@ -1111,19 +1128,55 @@ sub args_decode {
 }
 
 
+sub run_precheck {
+	my ($self) = @_;
+
+	for my $precheck (@{$self->{test}->{precheck}}) {
+		unless (system(@{$precheck}) == 0) {
+			$self->print_test_result('SKIP', "precheck failed");
+			$self->end_test('SKIP');
+		}
+	}
+
+	return 1;
+}
+
+
+sub find_program() {
+        my ($self, $pname) = @_;
+
+	my @directories = (".");
+	if ($self->{bin_sub_directory}) {
+	        push @directories, $self->{bin_sub_directory};
+	}
+
+	for my $up (('.', '..', '../..', '../../..')) {
+		for my $sub (('.', 'src')) {
+		        for my $dir (@directories) {
+				for my $ext (('', '.exe')) {
+					my $f = "$up/$sub/$dir/$pname$ext";
+					return $f if (-f $f);
+				}
+			}
+		}
+	}
+
+	return undef;
+}
+
+
 sub run_program {
 	my ($self) = @_;
-	goto &pipein_win32 if $^O eq 'MSWin32' && $self->{test}->{pipein};
 	my ($stdin, $stdout, $stderr);
 	$stderr = gensym;
 
-	my @cmd = ('../' . $self->{test}->{program}, map ({ args_decode($_, $self->{srcdir}); } @{$self->{test}->{args}}));
+	my @cmd = ($self->find_program($self->{test}->{program}), map ({ args_decode($_, $self->{srcdir}); } @{$self->{test}->{args}}));
 
 	### TODO: catch errors?
 
 	my $pid;
-        if ($self->{test}->{pipefile}) {
-                open(SPLAT, '<', $self->{test}->{pipefile});
+        if ($self->{test}->{'stdin-file'}) {
+                open(SPLAT, '<', $self->{test}->{'stdin-file'});
 	        my $is_marked = eof SPLAT; # mark used
 		$pid = open3("<&SPLAT", $stdout, $stderr, @cmd);
 	}
@@ -1133,39 +1186,22 @@ sub run_program {
 	$self->{stdout} = [];
 	$self->{stderr} = [];
 
-	if ($self->{test}->{pipein}) {
-                my $fh;
-                open($fh, "$self->{test}->{pipein} |");
-                if (!defined($fh)) {
-                        $self->die("cannot run pipein command [$self->{test}->{pipein}: $!");
+	if ($self->{test}->{stdin}) {
+		foreach my $line (@{$self->{test}->{stdin}}) {
+                        print $stdin $line . "\n";
                 }
-                while (my $line = <$fh>) {
-                        print $stdin $line;
-                }
-                close($fh);
                 close($stdin);
         }
 
 	while (my $line = <$stdout>) {
-		if ($^O eq 'MSWin32') {
-			$line =~ s/[\r\n]+$//;
-		}
-		else {
-			chomp $line;
-		}
+		$line =~ s/(\n|\r)//g;
 		push @{$self->{stdout}}, $line;
 	}
 	my $prg = $self->{test}->{program};
 	$prg =~ s,.*/,,;
 	while (my $line = <$stderr>) {
-		if ($^O eq 'MSWin32') {
-			$line =~ s/[\r\n]+$//;
-		}
-		else {
-			chomp $line;
-		}
-
-		$line =~ s/^[^: ]*$prg: //;
+		$line =~ s/(\n|\r)//g;
+		$line =~ s/^[^: ]*$prg(\.exe)?: //;
 		if (defined($self->{test}->{'stderr-replace'})) {
 			$line = $self->stderr_rewrite($self->{test}->{'stderr-replace'}, $line);
 		}
@@ -1177,40 +1213,6 @@ sub run_program {
 	$self->{exit_status} = $? >> 8;
 }
 
-sub pipein_win32() {
-	my ($self) = @_;
-
-	my $cmd = "$self->{test}->{pipein}| ..\\$self->{test}->{program} " . join(' ', map ({ args_decode($_, $self->{srcdir}); } @{$self->{test}->{args}}));
-	my ($success, $error_message, $full_buf, $stdout_buf, $stderr_buf) = IPC::Cmd::run(command => $cmd);
-	if (!$success) {
-		### TODO: catch errors?
-	}
-
-	my @stdout = map { s/[\r\n]+$// } @$stdout_buf;
-	$self->{stdout} = \@stdout;
-        $self->{stderr} = [];
-
-	my $prg = $self->{test}->{program};
-	$prg =~ s,.*/,,;
-	foreach my $line (@$stderr_buf) {
-		$line =~ s/[\r\n]+$//;
-
-		$line =~ s/^[^: ]*$prg: //;
-		if (defined($self->{test}->{'stderr-replace'})) {
-			$line = $self->stderr_rewrite($self->{test}->{'stderr-replace'}, $line);
-		}
-		push @{$self->{stderr}}, $line;
-	}
-
-	$self->{exit_status} = 1;
-	if ($success) {
-		$self->{exit_status} = 0;
-	}
-	elsif ($error_message =~ /exited with value ([0-9]+)$/) {
-		$self->{exit_status} = $1 + 0;
-	}
-}
-
 sub sandbox_create {
 	my ($self, $tag) = @_;
 
@@ -1252,10 +1254,9 @@ sub sandbox_leave {
 sub sandbox_remove {
 	my ($self) = @_;
 
-	my $ok = 1;
 	remove_tree($self->{sandbox_dir});
 
-	return $ok;
+	return 1;
 }
 
 
